@@ -38,9 +38,8 @@
 #' @export
 do_container_op <- function(container, operation="", options=list(), headers=list(), http_verb="GET", ...)
 {
-    # don't add trailing / if no within-container path supplied: ADLS will complain
     operation <- if(nchar(operation) > 0)
-        sub("//", "/", paste0(container$name, "/", operation))
+        paste0(container$name, "/", operation)
     else container$name
 
     call_storage_endpoint(container$endpoint, operation, options=options, headers=headers, http_verb=http_verb, ...)
@@ -57,7 +56,8 @@ call_storage_endpoint <- function(endpoint, path, options=list(), headers=list()
 {
     http_verb <- match.arg(http_verb)
     url <- httr::parse_url(endpoint$url)
-    url$path <- url_encode(path)
+    # fix doubled-up /'s which can result from file.path snafus etc
+    url$path <- gsub("/{2,}", "/", url_encode(file.path(url$path, path)))
     if(!is_empty(options))
         url$query <- options[order(names(options))] # must be sorted for access key signing
 
@@ -211,15 +211,26 @@ xml_to_list <- function(x)
 
 
 # check whether to retry a failed file transfer
-# retry on curl error (not any other kind of error)
-# don't retry on host not found
+# retry on:
+# - curl error (except host not found)
+# - http 400: MD5 mismatch
 retry_transfer <- function(res)
 {
-    inherits(res, "error") &&
-        grepl("curl", deparse(res$call[[1]]), fixed=TRUE) &&
+    UseMethod("retry_transfer")
+}
+
+retry_transfer.error <- function(res)
+{
+    grepl("curl", deparse(res$call[[1]]), fixed=TRUE) &&
         !grepl("Could not resolve host", res$message, fixed=TRUE)
 }
 
+retry_transfer.response <- function(res)
+{
+    httr::status_code(res) == 400L &&
+        grepl("Md5Mismatch", rawToChar(httr::content(res, as="raw")), fixed=TRUE)
+}
+
 
 as_datetime <- function(x, format="%a, %d %b %Y %H:%M:%S", tz="GMT")
 {
@@ -267,3 +278,9 @@ url_encode <- function(string, reserved=FALSE)
 {
     URLencode(enc2utf8(string), reserved=reserved)
 }
+
+
+encode_md5 <- function(x, ...)
+{
+    openssl::base64_encode(openssl::md5(x, ...))
+}
